from typing import Collection, Dict, FrozenSet, Iterator, Tuple, Optional

from fastani.config import VERSIONS
from fastani.logger import log


class FastANIParameters:
    __slots__ = ('version', 'exe', 'single_execution', 'k', 'cpus', 'frag_len',
                 'min_frac', 'min_frag', 'bidirectional')

    def __init__(self, version: int, exe: str,
                 single_execution: bool = True,
                 bidirectional: bool = False,
                 k: Optional[int] = None,
                 cpus: Optional[int] = None,
                 frag_len: Optional[int] = None,
                 min_frac: Optional[float] = None,
                 min_frag: Optional[int] = None):

        # Nullify parameters not specific to the current version
        if cpus is not None and version == VERSIONS.index('1.0'):
            cpus = None
            log.warning('Multi-threading is not supported in this version and will be ignored')
        if min_frac is not None and version < VERSIONS.index('1.3'):
            min_frac = None
            log.warning('Minimum fraction is not supported in this version and will be ignored')
        if min_frag is not None and version > VERSIONS.index('1.1 or 1.2'):
            min_frag = None
            log.warning('Minimum fragments is not supported in this version and will be ignored')

        # Validate the parameters
        cpus = max(1, cpus) if cpus is not None else None
        self.validate(k=k, frag_len=frag_len, min_frac=min_frac, min_frag=min_frag)

        # Set the parameters
        self.version = version
        self.exe = exe
        self.k = k
        self.cpus = cpus
        self.frag_len = frag_len
        self.min_frac = min_frac
        self.min_frag = min_frag

        self.single_execution = single_execution
        self.bidirectional = bidirectional

    @property
    def k_cmd(self):
        return ('-k', str(self.k)) if self.k is not None else tuple()

    @property
    def cpu_cmd(self):
        return ('-t', str(self.cpus)) if self.cpus is not None else tuple()

    @property
    def frag_len_cmd(self):
        return ('--fragLen', str(self.frag_len)) if self.frag_len is not None else tuple()

    @property
    def min_frac_cmd(self):
        return ('--minFraction', str(self.min_frac)) if self.min_frac is not None else tuple()

    @property
    def min_frag_cmd(self):
        return ('--minFrag', str(self.min_frag)) if self.min_frag is not None else tuple()

    @staticmethod
    def validate(k: Optional[int], frag_len: Optional[int],
                 min_frac: Optional[float], min_frag: Optional[int]):
        errors = list()
        # kmer size
        if k is not None and (k > 16 or k < 1):
            errors.append('kmer size must be between 1 and 16')
        # fragment length
        if frag_len is not None and frag_len < 1:
            errors.append('fragment length must be greater than 0')
        # minimum fraction size
        if min_frac is not None and (min_frac > 1.0 or min_frac < 0):
            errors.append('minimum fraction must be between 0 and 1')
        # minimum fragment size
        if min_frag is not None and min_frag < 0:
            errors.append('minimum fragments must be greater than 0')
        # report errors
        if len(errors) > 0:
            raise ValueError('\n'.join(errors))


class FastANIExecution:
    """The input, output and logs of a single FastANI execution."""

    __slots__ = ('cmd', 'stdout', 'stderr', 'return_code', 'output')

    def __init__(self, cmd: str, stdout: str, stderr: str, return_code: int,
                 output: str):
        self.cmd: str = cmd
        self.stdout: Optional[str] = stdout if stdout else None
        self.stderr: Optional[str] = stderr if stderr else None
        self.return_code: int = return_code
        self.output: Optional[ResultFile] = ResultFile(output) if output else None


class FastANIResult:
    __slots__ = ('ani', 'n_frag', 'total_frag', 'align_frac')

    def __init__(self, ani: float, n_frag: int, total_frag: int):
        self.ani: float = ani
        self.n_frag: int = n_frag
        self.total_frag: int = total_frag
        self.align_frac: float = n_frag / total_frag

    def __repr__(self):
        return f'ANI={self.ani}, AF={self.align_frac:.4f}'

    def __eq__(self, other):
        return isinstance(other, FastANIResult) and \
               self.ani == other.ani and \
               self.n_frag == other.n_frag and \
               self.total_frag == other.total_frag

    def to_delimited_str(self, delimiter: str = '\t') -> str:
        return delimiter.join(map(str, (self.ani, self.n_frag, self.total_frag,
                                        round(self.align_frac, 4))))


class FastANIResults:
    """The summary of a FastANI method request."""

    __slots__ = ('query', 'reference', 'params', 'executions')

    def __init__(self, query: FrozenSet[str], reference: FrozenSet[str],
                 executions: Tuple[FastANIExecution], params: FastANIParameters):
        self.query = query
        self.reference = reference
        self.params = params
        self.executions: Tuple[FastANIExecution] = executions

    def iter_results(self) -> Iterator[Tuple[str, str, Optional[FastANIResult]]]:
        """Returns an iterator of the query, reference, and results.
        NOTE: This will only return those which had >=80% ANI."""
        for execution in self.executions:
            if execution.output:
                for (qry, ref), result in execution.output.data.items():
                    yield qry, ref, result

    def as_dict(self) -> Dict[str, Dict[str, Optional[FastANIResult]]]:
        """Returns the results as a dictionary of query -> reference values."""

        # The output dictionary must be seeded with the query/reference genomes
        out = dict()
        for q in self.query:
            out[q] = dict()
            for r in self.reference:
                out[q][r] = None
                # Bidirectional case
                if self.params.bidirectional:
                    if r not in out:
                        out[r] = dict()
                    out[r][q] = None

        # Process the executions
        error = False
        for qry, ref, result in self.iter_results():
            out_result = out[qry][ref]
            if out_result is not None and out_result != result:
                error = True
                log.error(f'Duplicate and differing values, data is corrupt: {qry, ref, result, out_result}')
            else:
                out[qry][ref] = result
        if error:
            raise ValueError('Corrupt data')

        return out

    def to_file(self, path: str):
        with open(path, 'w') as f:
            for qry, ref_dict in sorted(self.as_dict().items(), key=lambda x: x[0]):
                for ref, result in sorted(ref_dict.items(), key=lambda x: x[0]):
                    cols = [qry, ref]
                    if result is not None:
                        cols.extend([result.ani, result.n_frag, result.total_frag,
                                     round(result.align_frac, 4)])
                    else:
                        cols.extend(['<80%', 'n/a', 'n/a', 'n/a'])
                    f.write('\t'.join(map(str, cols)) + '\n')
        log.info(f'Wrote results to {path}')


class ResultFile:
    """A wrapper to the FastANI output file."""

    __slots__ = ('data',)

    def __init__(self, path):
        self.data: Dict[Tuple[str, str], FastANIResult] = self.read(path)

    def __repr__(self):
        if len(self.data) == 0:
            return 'N/A'
        else:
            return f'{len(self.data):,} comparisons'

    @staticmethod
    def read(path):
        out = dict()
        with open(path) as f:
            for line in f.readlines():
                data = line.strip().split()
                if len(data) != 5:
                    raise Exception(f'Malformed output file: {line}')
                out[(data[0], data[1])] = FastANIResult(float(data[2]), int(data[3]), int(data[4]))
        return out


def write_genome_list(genomes: Collection[str], path: str):
    """Write a collection of genome names to the query/reference list file."""
    with open(path, 'w') as f:
        for genome_path in genomes:
            f.write(f'{genome_path}\n')



import os
import tempfile
from unittest import TestCase

from fastani import fastani
from test import GENOMES, KNOWN_ANI, EXE


class TestFastANI(TestCase):

    def test_execution_single_execution(self):
        query = GENOMES['a']
        reference = [GENOMES['b'], GENOMES['c'], GENOMES['d']]

        result = fastani(exe='fastANI_1.33', query=query, reference=reference, single_execution=True)
        self.assertEqual(len(result.executions), 1)

        result_dict = result.as_dict()

        self.assertEqual(len(result_dict), 1)
        self.assertEqual(len(result_dict[query]), len(reference))

        self.assertEqual(result_dict[query][GENOMES['b']], KNOWN_ANI['a']['b'])
        self.assertEqual(result_dict[query][GENOMES['c']], KNOWN_ANI['a']['c'])
        self.assertEqual(result_dict[query][GENOMES['d']], KNOWN_ANI['a']['d'])

    def test_execution_single_execution_bidirectional(self):
        query = GENOMES['a']
        reference = [GENOMES['b'], GENOMES['c'], GENOMES['d']]

        result = fastani(exe='fastANI_1.33', query=query, reference=reference, single_execution=True,
                         bidirectional=True)
        self.assertEqual(len(result.executions), 2)

        result_dict = result.as_dict()

        self.assertEqual(len(result_dict), 4)
        self.assertEqual(len(result_dict[query]), len(reference))

        self.assertEqual(result_dict[GENOMES['a']][GENOMES['b']], KNOWN_ANI['a']['b'])
        self.assertEqual(result_dict[GENOMES['a']][GENOMES['c']], KNOWN_ANI['a']['c'])
        self.assertEqual(result_dict[GENOMES['a']][GENOMES['d']], KNOWN_ANI['a']['d'])

        self.assertEqual(result_dict[GENOMES['b']][GENOMES['a']], KNOWN_ANI['b']['a'])
        self.assertEqual(result_dict[GENOMES['c']][GENOMES['a']], KNOWN_ANI['c']['a'])
        self.assertEqual(result_dict[GENOMES['d']][GENOMES['a']], KNOWN_ANI['d']['a'])

    def test_execution_multi(self):
        query = GENOMES['a']
        reference = [GENOMES['b'], GENOMES['c'], GENOMES['d']]

        result = fastani(exe='fastANI_1.33', query=query, reference=reference, single_execution=False)
        self.assertEqual(len(result.executions), 3)

        result_dict = result.as_dict()

        self.assertEqual(len(result_dict), 1)
        self.assertEqual(len(result_dict[query]), len(reference))

        self.assertEqual(result_dict[query][GENOMES['b']], KNOWN_ANI['a']['b'])
        self.assertEqual(result_dict[query][GENOMES['c']], KNOWN_ANI['a']['c'])
        self.assertEqual(result_dict[query][GENOMES['d']], KNOWN_ANI['a']['d'])

    def test_execution_multi_bidirectional(self):
        query = GENOMES['a']
        reference = [GENOMES['b'], GENOMES['c'], GENOMES['d']]

        result = fastani(exe='fastANI_1.33', query=query, reference=reference, single_execution=False,
                         bidirectional=True)
        self.assertEqual(len(result.executions), 6)

        result_dict = result.as_dict()

        self.assertEqual(len(result_dict), 4)
        self.assertEqual(len(result_dict[query]), len(reference))

        self.assertEqual(result_dict[GENOMES['a']][GENOMES['b']], KNOWN_ANI['a']['b'])
        self.assertEqual(result_dict[GENOMES['a']][GENOMES['c']], KNOWN_ANI['a']['c'])
        self.assertEqual(result_dict[GENOMES['a']][GENOMES['d']], KNOWN_ANI['a']['d'])

        self.assertEqual(result_dict[GENOMES['b']][GENOMES['a']], KNOWN_ANI['b']['a'])
        self.assertEqual(result_dict[GENOMES['c']][GENOMES['a']], KNOWN_ANI['c']['a'])
        self.assertEqual(result_dict[GENOMES['d']][GENOMES['a']], KNOWN_ANI['d']['a'])

    def test_fastani_multi_version(self):
        query = GENOMES['a']
        reference = GENOMES['b']

        for exe in EXE:
            result = fastani(exe=exe, query=query, reference=reference, single_execution=True)
            self.assertEqual(len(result.executions), 1)

    def test_fastani_multiprocessing(self):
        query = GENOMES['a']
        reference = [GENOMES['b'], GENOMES['c'], GENOMES['d']]

        result = fastani(exe='fastANI_1.33', query=query, reference=reference,
                         single_execution=False, bidirectional=True, cpus=4)

        result_dict = result.as_dict()
        self.assertEqual(len(result.executions), 6)

        self.assertEqual(len(result_dict), 4)
        self.assertEqual(len(result_dict[query]), len(reference))

        self.assertEqual(result_dict[GENOMES['a']][GENOMES['b']], KNOWN_ANI['a']['b'])
        self.assertEqual(result_dict[GENOMES['a']][GENOMES['c']], KNOWN_ANI['a']['c'])
        self.assertEqual(result_dict[GENOMES['a']][GENOMES['d']], KNOWN_ANI['a']['d'])

        self.assertEqual(result_dict[GENOMES['b']][GENOMES['a']], KNOWN_ANI['b']['a'])
        self.assertEqual(result_dict[GENOMES['c']][GENOMES['a']], KNOWN_ANI['c']['a'])
        self.assertEqual(result_dict[GENOMES['d']][GENOMES['a']], KNOWN_ANI['d']['a'])

    def test_fastani_missing_path(self):
        query = GENOMES['a']
        reference = '/tmp/not-exist'

        for exe in EXE:
            self.assertRaises(FileNotFoundError, fastani, exe=exe, query=query, reference=reference,
                              single_execution=False)
            self.assertRaises(FileNotFoundError, fastani, exe=exe, query=query, reference=reference,
                              single_execution=True)

    def test_fastani_diff_parameters(self):
        query = GENOMES['a']
        reference = GENOMES['b']

        result = fastani(exe='fastANI_1.33', query=query, reference=reference,
                         k=10, frag_len=1000, min_frac=0.7,
                         single_execution=True, bidirectional=False)

        self.assertTrue('-k 10' in result.executions[0].cmd)
        self.assertTrue('--fragLen 1000' in result.executions[0].cmd)
        self.assertTrue('--minFraction 0.7' in result.executions[0].cmd)

    def test_fastani_write_results_to_disk(self):
        query = [GENOMES['a'], GENOMES['b'], GENOMES['c'], GENOMES['d']]
        reference = query

        result = fastani(exe='fastANI_1.33', query=query, reference=reference,
                         single_execution=True, bidirectional=False, cpus=8)

        with tempfile.TemporaryDirectory() as tmpdir:
            path_out = os.path.join(tmpdir, 'results.txt')

            result.to_file(path_out)

            with open(path_out, 'r') as f:
                data = f.read()

            self.assertEqual(len(data.splitlines()), 16)



from typing import Optional, Collection, Union

from fastani.config import VERSIONS
from fastani.model import FastANIParameters, FastANIResults
from fastani.process import process_queue
from fastani.util import validate_paths
from fastani.version import get_fastani_version


def fastani(query: Union[str, Collection[str]],
            reference: Union[str, Collection[str]],
            k: Optional[int] = None,
            cpus: Optional[int] = None,
            frag_len: Optional[int] = None,
            min_frac: Optional[float] = None,
            min_frag: Optional[int] = None,
            single_execution: bool = True,
            bidirectional: bool = False,
            exe: str = 'fastANI') -> FastANIResults:
    """
    :param query: Path to the query genome(s).
    :param reference: Path to the reference genome(s).
    :param k: kmer size <= 16 (default: 16)
    :param cpus: Number of CPUs to use for parallel execution (default: 1)
    :param frag_len: fragment length (default: 3,000)
    :param min_frac: minimum fraction of genome that must be shared for
                     trusting ANI (default: 0.2) [version >= 1.3]
    :param min_frag: minimum matched fragments for trusting ANI (default: 50)
                     [version <= 1.2].
    :param single_execution: True if --refList and --queryList should be
                  used, otherwise a subprocess will be launched to do 1:1
                  comparisons.
    :param bidirectional: True if the ANI should be calculated for query
                          vs reference and vice-versa.
    :param exe: The path to the execution file.
    """

    # Transform input to a collection
    query = validate_paths(query)
    reference = validate_paths(reference)

    # Nullify version-specific arguments
    version = VERSIONS.index(get_fastani_version(exe))
    params = FastANIParameters(version=version, exe=exe,
                               k=k, cpus=cpus, frag_len=frag_len,
                               min_frac=min_frac, min_frag=min_frag,
                               single_execution=single_execution,
                               bidirectional=bidirectional)

    # Create the queue
    queue = list()

    # Run the single execution (query/reference lists) and return the results
    if params.single_execution:
        queue.append({'ql': query, 'rl': reference})
        if params.bidirectional:
            queue.append({'ql': reference, 'rl': query})

    # Generate a queue of commands
    else:
        for query_path in query:
            for reference_path in reference:
                queue.append({'q': query_path, 'r': reference_path})
                if params.bidirectional:
                    queue.append({'q': reference_path, 'r': query_path})

    # Process the queue and return the result
    executions = process_queue(params, queue)
    return FastANIResults(query=query, reference=reference, executions=executions, params=params)



from unittest import TestCase

from fastani.model import FastANIResult, FastANIResults
from fastani.util import validate_paths


class TestModel(TestCase):

    def test_fastani_result_eq(self):
        a = FastANIResult(ani=89.1234, n_frag=123, total_frag=444)
        b = FastANIResult(ani=89.1234, n_frag=123, total_frag=444)
        c = FastANIResult(ani=89.1234, n_frag=222, total_frag=444)

        self.assertEqual(a, b)
        self.assertNotEqual(a, c)





import subprocess


def get_fastani_version(exe: str) -> str:
    """Determines the version of FastANI based on the executable help text."""

    proc = subprocess.Popen([exe, '--version'], stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, encoding='utf-8')
    stdout, stderr = proc.communicate()

    # Not supported, perhaps it's an older version
    if proc.returncode != 0:
        proc = subprocess.Popen([exe, '--help'], stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        if proc.returncode != 0 and not 'fastANI is a fast alignment-free' in stderr:
            raise Exception('Could not determine FastANI version')
        if '--matrix' in stderr:
            return '1.1 or 1.2'
        else:
            return '1.0'
    else:
        return stderr.strip().split()[-1]



import os
import re
import subprocess
import tempfile
from multiprocessing import Pool
from typing import Optional, Collection, Tuple, Dict

from fastani.model import FastANIParameters, FastANIExecution
from fastani.model import write_genome_list

re_missing_file = re.compile(r'ERROR, skch::validateInputFiles, Could not open (.+)')


def execute_mp_wrapper(args: Tuple[FastANIParameters, Dict[str, str]]) -> FastANIExecution:
    """Required as multiprocessing pool will only allow one argument."""
    return execute(params=args[0], **args[1])


def execute(params: FastANIParameters, q: Optional[str] = None, ql: Optional[Collection[str]] = None,
            r: Optional[str] = None, rl: Optional[Collection[str]] = None) -> FastANIExecution:
    cmd = [params.exe]
    if q is not None:
        cmd.extend(['--query', q])
    if r is not None:
        cmd.extend(['--ref', r])
    cmd.extend(params.k_cmd)
    cmd.extend(params.cpu_cmd)
    cmd.extend(params.frag_len_cmd)
    cmd.extend(params.min_frac_cmd)
    cmd.extend(params.min_frag_cmd)

    with tempfile.TemporaryDirectory(prefix='fastani_') as temp_dir:

        # Set a temporary output file path if output is not set
        out_path = os.path.join(temp_dir, 'output.txt')
        cmd.extend(['--output', out_path])

        # Create a temporary file for the query/ref lists and write to disk
        if ql is not None:
            path_ql = os.path.join(temp_dir, 'query.txt')
            write_genome_list(ql, path_ql)
            cmd.extend(['--queryList', path_ql])
        if rl is not None:
            path_rl = os.path.join(temp_dir, 'reference.txt')
            write_genome_list(rl, path_rl)
            cmd.extend(['--refList', path_rl])

        # Run the proc
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        if proc.returncode != 0:
            missing_file = re_missing_file.search(stderr)
            if missing_file:
                raise FileNotFoundError(missing_file.group(1))
            raise Exception(f'Failed to run {params.exe} with error: {stderr}')

        # Package the results and return them
        return FastANIExecution(' '.join(cmd), stdout, stderr, proc.returncode, output=out_path)


def process_queue(params: FastANIParameters, queue) -> Tuple[FastANIExecution]:
    # If only using one CPU, don't create a multiprocessing pool
    if params.cpus == 1 or len(queue) <= 1 or params.single_execution:
        return tuple([execute(params=params, **cmd) for cmd in queue])

    # Do single executions, i.e. force FastANI execution to be single threaded
    else:
        mp_cpus = params.cpus
        params.cpus = 1
        mp_queue = [(params, item) for item in queue]
        with Pool(processes=mp_cpus) as pool:
            return tuple(pool.imap_unordered(execute_mp_wrapper, mp_queue))
